/*
 * Copyright 2013 Netflix, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package feign.codec;

import feign.Response;
import org.testng.annotations.Test;
import org.w3c.dom.Document;

import java.io.ByteArrayInputStream;
import java.io.InputStream;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;

import static org.testng.Assert.assertEquals;
import static org.testng.Assert.assertNull;

import static feign.Util.UTF_8;

public class DefaultDecoderTest {
  private final Decoder decoder = new Decoder.Default();

  @Test public void testDecodesToString() throws Exception {
    Response response = knownResponse();
    Object decodedObject = decoder.decode(response, String.class);
    assertEquals(decodedObject.getClass(), String.class);
    assertEquals(decodedObject.toString(), "response body");
  }

  @Test public void testDecodesToByteArray() throws Exception {
    Response response = knownResponse();
    Object decodedObject = decoder.decode(response, byte[].class);
    assertEquals(decodedObject.getClass(), byte[].class);
    assertEquals((byte[]) decodedObject, "response body".getBytes(UTF_8));
  }

  @Test public void testDecodesNullBodyToNull() throws Exception {
    assertNull(decoder.decode(nullBodyResponse(), Document.class));
  }

  @Test(expectedExceptions = DecodeException.class, expectedExceptionsMessageRegExp = ".* is not a type supported by this decoder.")
  public void testRefusesToDecodeOtherTypes() throws Exception {
    decoder.decode(knownResponse(), Document.class);
  }

  private Response knownResponse() {
    String content = "response body";
    InputStream inputStream = new ByteArrayInputStream(content.getBytes(UTF_8));
    Map<String, Collection<String>> headers = new HashMap<String, Collection<String>>();
    headers.put("Content-Type", Collections.singleton("text/plain"));
    return Response.create(200, "OK", headers, inputStream, content.length());
  }

  private Response nullBodyResponse() {
    return Response.create(200, "OK", Collections.<String, Collection<String>>emptyMap(), null);
  }
}
